import io
import pytest
import itertools
from pathlib import Path
from setuptools_github import tools


class REFilter:
    def __init__(self, expr):
        from re import compile

        self.expr = compile(expr)

    def __repr__(self):
        return self.expr.pattern


class RESearch(REFilter):
    "build an object with a .search method, useable to filter"

    def __init__(self, expr, invert=False):
        super().__init__(expr)
        self.invert = invert

    def search(self, txt):
        out = self.expr.search(txt)
        return (not out) if self.invert else out


class REMatch(REFilter):
    def __eq__(self, actual):
        return bool(self.expr.match(actual))


# this is the output from ${{ toJson(github) }}
GITHUB = {
    "beta": {
        "ref": "refs/heads/beta/0.0.4",
        "sha": "2169f90c22e",
        "run_number": "8",
    },
    "release": {
        "ref": "refs/tags/release/0.0.3",
        "sha": "5547365c82",
        "run_number": "3",
    },
    "master": {
        "ref": "refs/heads/master",
        "sha": "2169f90c",
        "run_number": "20",
    },
}


def test_abort_exception():
    a = tools.AbortExecution(
        "a one-line error message",
        """
        A multi line
          explaination of
           what happened
         with some detail
    """,
        """
    Another multiline hint how
      to fix the issue
    """,
    )

    assert a.message == "a one-line error message"
    assert (
        f"\n{a.explain}\n"
        == """
A multi line
  explaination of
   what happened
 with some detail
"""
    )
    assert (
        f"\n{a.hint}\n"
        == """
Another multiline hint how
  to fix the issue
"""
    )

    assert (
        f"\n{str(a)}\n"
        == """
a one-line error message
  A multi line
    explaination of
     what happened
   with some detail
hint:
  Another multiline hint how
    to fix the issue
"""
    )

    a = tools.AbortExecution("hello world")
    assert a.message == "hello world"
    assert a.explain == ""
    assert a.hint == ""
    assert str(a) == "hello world"


def test_indent():
    txt = """
    This is a simply
       indented text
      with some special
         formatting
"""
    expected = """
..This is a simply
..   indented text
..  with some special
..     formatting
"""

    found = tools.indent(txt[1:], "..")
    assert f"\n{found}" == expected


def test_hubversion():
    "extracts from a GITHUB a (version, hash) tuple"

    pytest.raises(
        tools.InvalidGithubReference,
        tools.hubversion,
        {"ref": "", "run_number": "", "sha": ""},
        "",
    )

    fallbacks = [
        "123",
        "",
    ]

    expects = {
        ("beta", ""): ("0.0.4b8", "2169f90c22e"),
        ("beta", "123"): ("0.0.4b8", "2169f90c22e"),
        ("release", "123"): ("0.0.3", "5547365c82"),
        ("release", ""): ("0.0.3", "5547365c82"),
        ("master", "123"): ("123", "2169f90c"),
        ("master", ""): ("", "2169f90c"),
    }

    itrange = itertools.product(GITHUB, fallbacks)
    for key, fallback in itrange:
        gdata = GITHUB[key]
        expected = expects[(key, fallback)]
        assert expected == tools.hubversion(gdata, fallback)


def test_get_module_var(tmp_path):
    "pulls variables from a file"
    path = tmp_path / "in0.txt"
    path.write_text(
        """
# a test file
A = 12
B = 3+5
C = "hello"
# end of test
"""
    )
    assert 12 == tools.get_module_var(path, "A")
    assert "hello" == tools.get_module_var(path, "C")
    pytest.raises(AssertionError, tools.get_module_var, path, "B")
    pytest.raises(tools.MissingVariable, tools.get_module_var, path, "X1")


def test_set_module_var_empty_file(tmp_path):
    "check if the set_module_var will create a bew file"
    path = tmp_path / "in1.txt"

    assert not path.exists()
    tools.set_module_var(path, "__version__", "1.2.3")

    assert path.exists()
    path.write_text("# a fist comment line\n" + path.read_text().strip())

    tools.set_module_var(path, "__hash__", "4.5.6")
    assert (
        path.read_text().strip()
        == """
# a fist comment line
__version__ = "1.2.3"
__hash__ = "4.5.6"
""".strip()
    )


def test_set_module_var(tmp_path):
    "handles set_module_var cases"
    path = tmp_path / "in2.txt"

    path.write_text(
        """
# a fist comment line
__hash__ = "4.5.6"
# end of test
"""
    )

    version, txt = tools.set_module_var(path, "__version__", "1.2.3")
    assert not version
    assert (
        txt.rstrip()
        == """
# a fist comment line
__hash__ = "4.5.6"
# end of test
__version__ = "1.2.3"
""".rstrip()
    )

    version, txt = tools.set_module_var(path, "__version__", "6.7.8")
    assert version == "1.2.3"
    assert (
        txt.rstrip()
        == """
# a fist comment line
__hash__ = "4.5.6"
# end of test
__version__ = "6.7.8"
""".rstrip()
    )

    version, txt = tools.set_module_var(path, "__hash__", "9.10.11")
    assert version == "4.5.6"
    assert (
        txt.rstrip()
        == """
# a fist comment line
__hash__ = "9.10.11"
# end of test
__version__ = "6.7.8"
""".rstrip()
    )
    return


def test_update_version(tmp_path):
    "test the update_version processing"
    from hashlib import sha224

    def writeinit(path):
        path.write_text(
            """
# a test file
__version__ = "1.2.3"
__hash__ = "4.5.6"

# end of test
"""
        )
        return sha224(path.read_bytes()).hexdigest()

    initfile = tmp_path / "__init__.py"
    hashval = writeinit(initfile)

    # verify nothing has changed
    assert "1.2.3" == tools.update_version(initfile)
    assert hashval == sha224(initfile.read_bytes()).hexdigest()

    # we update the __version__/__hash__ from a master branch
    tools.update_version(initfile, GITHUB["master"])
    assert (
        initfile.read_text()
        == """
# a test file
__version__ = "1.2.3"
__hash__ = "2169f90c"

# end of test
"""
    )

    # we update __version__/__hash__ from a beta branch (note the b<build-number>)
    writeinit(initfile)
    tools.update_version(initfile, GITHUB["beta"])
    assert (
        initfile.read_text()
        == """
# a test file
__version__ = "0.0.4b8"
__hash__ = "2169f90c22e"

# end of test
"""
    )

    writeinit(initfile)
    tools.update_version(initfile, GITHUB["release"])
    assert (
        initfile.read_text()
        == """
# a test file
__version__ = "0.0.3"
__hash__ = "5547365c82"

# end of test
"""
    )


def test_bump_version():
    "bump version test"
    assert tools.bump_version("0.0.1", "micro") == "0.0.2"
    assert tools.bump_version("0.0.2", "micro") == "0.0.3"
    assert tools.bump_version("0.0.2", "minor") == "0.1.0"
    assert tools.bump_version("1.2.3", "major") == "2.0.0"
    assert tools.bump_version("1.2.3", "release") == "1.2.3"


@pytest.mark.parametrize("method", ["no-force-no-keep", "no-force-keep"])
def test_gitwrapper_init(tmp_path, method):
    repo = tools.GitWrapper(tmp_path / "wow")
    assert repo.workdir == tmp_path / "wow"
    assert not repo.workdir.exists()

    def no_force_no_keep():
        repo.init(force=False, keepfile=False)
        assert repo.workdir.exists()
        assert not [
            p
            for p in repo.workdir.rglob("*")
            if not str(p.relative_to(repo.workdir)).startswith(".git")
        ]
        pytest.raises(FileExistsError, repo.init, force=False, keepfile=False)

    def no_force_keep():
        repo.init(force=False, keepfile=True)
        assert repo.workdir.exists()
        assert [
            p.relative_to(repo.workdir)
            for p in repo.workdir.rglob("*")
            if not str(p.relative_to(repo.workdir)).startswith(".git")
        ] == [Path(".keep")]
        pytest.raises(FileExistsError, repo.init, force=False, keepfile=True)

    locals()[method.replace("-", "_")]()
    assert repo.workdir == tmp_path / "wow"
    assert repo.workdir.exists()


def test_gitwrapper_clone(tmp_path):
    repo = tools.GitWrapper(tmp_path / "repo").init()
    assert repo(["rev-parse", "--abbrev-ref", "HEAD"]).strip() == "master"

    repo(["checkout", "-b", "a-branch"])
    assert repo(["rev-parse", "--abbrev-ref", "HEAD"]).strip() == "a-branch"

    # clone into project
    project = repo.clone(tmp_path / "project", force=False)
    assert project(["rev-parse", "--abbrev-ref", "HEAD"]).strip() == "a-branch"

    # clone without and with the force flag
    pytest.raises(ValueError, repo.clone, tmp_path / "project", force=False)
    repo.clone(tmp_path / "project", force=True)

    # re-clone the master branch
    project = repo.clone(tmp_path / "project", branch="master", force=True)
    assert project(["rev-parse", "--abbrev-ref", "HEAD"]).strip() == "master"


def test_gitwrapper_dump(tmp_path):
    repo = tools.GitWrapper(tmp_path / "repo2").init()

    # adds a file
    path = repo / "abc.txt"
    path.write_text("hello")
    repo(["add", path])
    repo(["commit", "-m", "initial", path])
    assert repo(["rev-parse", "--abbrev-ref", "HEAD"]).strip() == "master"

    # clone the project
    project = repo.clone(tmp_path / "wow-cloned", force=True)

    # check the output
    expected = REMatch(
        r"""
REPO: .*
 .status.
  On branch master
  Your branch is up to date with 'origin/master'.
\s+
  nothing to commit, working tree clean

 .branch.
  . master               \s+[0-9a-f]{7} .origin/master. initial
    remotes/origin/HEAD  \s+-> origin/master
    remotes/origin/master\s+[0-9a-f]{7} initial

 .tags.
\s+
 .remote.
  origin\s.*[(]fetch[)]
  origin\s.*[(]push[)]
""".strip()
    )
    assert project.dump(io.StringIO) == expected


def test_gitcli(tmp_path):
    repo = tools.GitCli(tmp_path / "wow1").init(force=True)
    assert repo.branch() == "master"
    assert (repo.branch(name="opla"), repo.branch()) == ("master", "opla")
    assert repo.branches() == (["master", "opla"], {})

    project = repo.clone(tmp_path / "wow1-cloned", force=True)
    assert project.branch() == "opla"

    project = repo.clone(tmp_path / "wow1-cloned", branch="master", force=True)
    assert project.branch() == "master"
    assert (project.branch(name="wow"), project.branch()) == ("master", "wow")

    assert project.branches() == (
        ["master", "wow"],
        {"origin": ["HEAD", "master", "opla"]},
    )
    assert project.branches(RESearch("[A-Z]", True)) == (
        ["master", "wow"],
        {"origin": ["master", "opla"]},
    )

    path = project.workdir / "abc.txt"
    path.write_text("hello")
    project.commit(path, "a commit")



import re
from unittest import mock
import pytest
import contextlib
import argparse

from setuptools_github import start_release as sr
from setuptools_github import tools


@pytest.fixture(scope="function")
def cli():
    from argparse import Namespace

    with contextlib.ExitStack() as stack:
        mcks = {}
        mcks["release"] = stack.enter_context(
            mock.patch("setuptools_github.start_release.release")
        )
        mcks["beta"] = stack.enter_context(
            mock.patch("setuptools_github.start_release.beta")
        )

        mcks["exit"] = stack.enter_context(
            mock.patch.object(argparse.ArgumentParser, "exit")
        )
        mcks["error"] = stack.enter_context(
            mock.patch.object(argparse.ArgumentParser, "error")
        )
        mcks["_print_message"] = stack.enter_context(
            mock.patch.object(argparse.ArgumentParser, "_print_message")
        )
        yield Namespace(**mcks)


def test_extract_beta_branches(git_project_factory):
    "test the branch and tag extraction function"
    from pygit2 import Repository

    repo = git_project_factory("test_check_version-repo").create("0.0.0")
    repo1 = git_project_factory("test_check_version-repo1").create(clone=repo)

    repo.branch("beta/0.0.3")
    repo(["tag", "-m", "release", "release/0.0.3"])
    repo.branch("beta/0.0.4")
    repo(["tag", "-m", "release", "release/0.0.4"])
    repo1.branch("beta/0.0.2")

    project = git_project_factory().create(clone=repo)
    project.branch("beta/0.0.1", "origin/master")
    project.branch("master", "origin/master")

    project(["remote", "add", "repo1", repo1.workdir])
    project(["fetch", "--all"])

    local_branches, remote_branches, tags = [
        *project.branches(project.BETA_BRANCHES),
        project(["tag", "-l"]).split(),
    ]

    repo = Repository(project.workdir)
    assert (local_branches, remote_branches, tags) == sr.extract_beta_branches(repo)

    assert local_branches == ["beta/0.0.1", "beta/0.0.4"]
    assert remote_branches == {
        "origin": ["beta/0.0.3", "beta/0.0.4"],
        "repo1": ["beta/0.0.2"],
    }
    assert tags == ["release/0.0.3", "release/0.0.4"]


def test_start_release_repo_has_modifications(cli, git_project_factory):
    "verifies there are no local modified files"
    project = git_project_factory().create()

    # no init file at all -> fail
    options = sr.parse_args(
        ["-w", str(project.workdir), "micro", str(project.initfile)], testmode=True
    )
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "no init file found"
    assert (cli.release.called, cli.beta.called) == (False, False)

    # add initfile, but not commit -> fail
    project.initfile.parent.mkdir(parents=True, exist_ok=True)
    project.initfile.write_text("__version__ = '0.0.0'")
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "init file is not tracked"
    assert (cli.release.called, cli.beta.called) == (False, False)

    # modify committed initfile -> fail
    project(["add", project.initfile])
    project(["commit", "-m", "test", project.initfile])
    project.initfile.write_text("__version__ = '0.0.0'\n")
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "init file has local modifications"
    assert (cli.release.called, cli.beta.called) == (False, False)


def test_start_release_invalid_init(cli, git_project_factory):
    "check the initfile contains at lease the __version__ information"
    project = git_project_factory().create()
    options = sr.parse_args(
        ["-w", str(project.workdir), "micro", str(project.initfile)], testmode=True
    )

    # no init file at all
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "no init file found"
    assert (cli.release.called, cli.beta.called) == (False, False)

    # adds an invalid init file
    project.initfile.parent.mkdir(parents=True, exist_ok=True)
    project.initfile.write_text("hello")
    project(["add", project.initfile])
    project(["commit", "-m", "test", project.initfile])

    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "init file has an invalid __version__ module variable"
    assert (cli.release.called, cli.beta.called) == (False, False)


def test_start_release_invalid_branch(cli, git_project_factory):
    "verify the starting branch for beta and releases"
    project = git_project_factory().create("0.0.0")
    options = sr.parse_args(
        ["-w", str(project.workdir), "micro", str(project.initfile)], testmode=True
    )

    project.branch("abc")
    project(["checkout", "abc"])
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "'micro' starts from 'master' branch"

    options = sr.parse_args(
        ["-w", str(project.workdir), "release", str(project.initfile)], testmode=True
    )
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "release starts from a beta/N.M.O branch"


def test_start_release_invalid_version(cli, git_project_factory):
    "check for beta/release colliding versions"
    repo = git_project_factory("test_start_release_invalid_version-repo").create(
        "0.0.0"
    )
    repo1 = git_project_factory("test_start_release_invalid_version-repo1").create(
        clone=repo
    )

    repo.branch("beta/0.0.3")
    repo(["tag", "-m", "release", "release/0.0.3"])
    repo.branch("beta/0.0.4")
    repo(["tag", "-m", "release", "release/0.0.4"])
    repo1.branch("beta/0.0.2")

    project = git_project_factory().create(clone=repo)
    project.branch("beta/0.0.1", "origin/master")
    project.branch("master", "origin/master")

    project(["remote", "add", "repo1", repo1.workdir])
    project(["fetch", "--all"])

    assert project.branch(), project.version == ("master", "0.0.0")

    # cannot re-create a "local" branch
    options = sr.parse_args(
        ["-w", str(project.workdir), "micro", str(project.initfile)], testmode=True
    )
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert (
        exc.message
        == "next version branch 'beta/0.0.1' already present in local branches"
    )

    # cannot re-create a "remote" branch
    project.initfile.write_text("__version__ = '0.0.1'")
    project.commit(project.initfile, "update")
    assert project.branch(), project.version == ("master", "0.0.1")
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert (
        exc.message
        == "next version branch 'beta/0.0.2' already present in remote branches"
    )

    project(["checkout", "beta/0.0.4"])
    assert project.branch(), project.version == ("beta/0.0.4", "0.0.0")

    options = sr.parse_args(
        ["-w", str(project.workdir), "release", str(project.initfile)], testmode=True
    )
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "wrong version '0.0.0' from initfile"


def test_start_cli_help(cli):
    "test the --help flag to the cli"
    sr.parse_args(["--help"], testmode=True)

    # we show only the help
    assert (cli.release.called, cli.beta.called) == (False, False)

    def s(txt):
        result = re.sub(r"\s(\s*)", " ", txt.strip())
        if ".py optional arguments:" in result:
            result = result.replace(".py optional arguments:", ".py options:")
        return result

    assert s(cli._print_message.call_args[0][0]) == s(
        """
usage: AAAA [-h] [-n] [-v] [--master MASTER] [-w WORKDIR]
                            {micro,minor,major,release} __init__.py

positional arguments:
  {micro,minor,major,release}
  __init__.py

options:
  -h, --help            show this help message and exit
  -n, --dry-run
  -v, --verbose
  --master MASTER       the 'master' branch (default: master)
  -w WORKDIR, --workdir WORKDIR
                        git working dir (default: .)
"""
    )


def test_start_release_safe_beta_cli(cli, git_project_factory):
    "test the cli usage for beta releases up to the beta execution"
    project = git_project_factory("test_start_release_beta_repo").create()

    # we start with an empty repo
    options = sr.parse_args(
        ["-w", str(project.workdir), "micro", str(project.initfile)], testmode=True
    )
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "no init file found"

    # we add an (invalid) init file
    project.initfile.parent.mkdir(parents=True, exist_ok=True)
    project.initfile.write_text("hello")
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "init file is not tracked"

    # we track it (but no commit)
    project(["add", project.initfile])
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "init file has local modifications"

    # we commit changes (still has no __version__ in it)
    project(["commit", "-m", "initial"])
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "init file has an invalid __version__ module variable"

    # ok we fix the __version__
    project.initfile.write_text("__version__ = '0.0.0'")
    project(["commit", "-m", "start", project.initfile])

    # and we verify we're able to call the beta function
    assert (cli.release.called, cli.beta.called) == (False, False)
    sr.run(**options)
    assert (cli.release.called, cli.beta.called) == (False, True)


def test_start_release_safe_release_cli(cli, git_project_factory):
    "test the cli usage for a release up to the release execution"

    project = git_project_factory().create("0.0.0")
    assert (project.branch(), project.version) == ("master", "0.0.0")

    # we start on master (eg. we cannot start a release)
    options = sr.parse_args(
        ["-w", str(project.workdir), "release", str(project.initfile)], testmode=True
    )
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "release starts from a beta/N.M.O branch"

    # ok we can start a beta release here
    options = sr.parse_args(
        ["-w", str(project.workdir), "micro", str(project.initfile)], testmode=True
    )
    sr.run(**options)
    assert (cli.release.called, cli.beta.called) == (False, True)
    cli.beta.reset_mock()

    # we start a release from a named branch
    project.branch("beta/0.0.0")
    assert (project.branch(), project.version) == ("beta/0.0.0", "0.0.0")
    options = sr.parse_args(
        ["-w", str(project.workdir), "release", str(project.initfile)], testmode=True
    )
    sr.run(**options)
    assert (cli.release.called, cli.beta.called) == (True, False)


def test_start_release_e2e(git_project_factory):
    "full end2end walkthrough scenario"
    project = git_project_factory().create("0.0.1")
    assert (project.branch(), project.version) == ("master", "0.0.1")

    # creating a first branch (note as we won't bump 0.0.1 because is
    # the first beta branch)
    options = sr.parse_args(
        ["-w", str(project.workdir), "micro", str(project.initfile)], testmode=True
    )
    sr.run(**options)
    assert (project.branch(), project.version) == ("beta/0.0.1", "0.0.1")

    # try a beta release from a beta branch should fail
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "'micro' starts from 'master' branch"

    # trying a second branch
    project(["checkout", "master"])
    assert (project.branch(), project.version) == ("master", "0.0.1")

    options = sr.parse_args(
        ["-w", str(project.workdir), "micro", str(project.initfile)], testmode=True
    )
    sr.run(**options)
    assert (project.branch(), project.version) == ("beta/0.0.2", "0.0.2")

    # again we won't be able to start a new branch from a non master branch
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "'micro' starts from 'master' branch"

    # two more betas
    project(["checkout", "master"])
    options = sr.parse_args(
        ["-w", str(project.workdir), "micro", str(project.initfile)], testmode=True
    )
    sr.run(**options)
    assert (project.branch(), project.version) == ("beta/0.0.3", "0.0.3")

    project(["checkout", "master"])
    options = sr.parse_args(
        ["-w", str(project.workdir), "micro", str(project.initfile)], testmode=True
    )
    sr.run(**options)
    assert (project.branch(), project.version) == ("beta/0.0.4", "0.0.4")

    # let's finally release 0.0.3 here (first we verify we cannot release
    # from a non master branch)
    project(["checkout", "beta/0.0.3"])
    exc = pytest.raises(tools.AbortExecution, sr.run, **options).value
    assert exc.message == "'micro' starts from 'master' branch"

    assert not project(["tag", "-l"]).split()
    options = sr.parse_args(
        ["-w", str(project.workdir), "release", str(project.initfile)], testmode=True
    )
    sr.run(**options)
    assert project(["tag", "-l"]).split() == ["release/0.0.3"]



from __future__ import annotations
import os
import io
import re
import sys
import pathlib
import shutil
import contextlib
import collections
import subprocess
from typing import Union, Optional, Tuple, TextIO, List, Dict, Any

import pytest


@pytest.fixture()
def datadir(request):
    basedir = pathlib.Path(__file__).parent / "data"
    if os.getenv("DATADIR"):
        basedir = pathlib.Path(os.getenv("DATADIR"))
    basedir = basedir / getattr(request.module, "DATADIR", "")
    return basedir


@pytest.fixture()
def scripter(request, tmp_path_factory, datadir):
    """handles script (cli) execution

    def test(scripter):
        script = scripter / "script-file.py"
        result = script.run(["--help"]) # this will execute:
                                        #   script-file.py --help
        assert result.out and result.err
    """
    Result = collections.namedtuple("R", "out,err,code")

    class ScripterError(Exception):
        pass

    class MissingItemError(ScripterError):
        pass

    class Exe:
        def __repr__(self):
            return (
                f"<{self.__class__.__name__} script={self.script} at {hex(id(self))}>"
            )

        def __init__(self, script, workdir, datadir, exe):
            self.script = script
            self.workdir = workdir
            self.datadir = datadir
            self.exe = exe
            if not pathlib.Path(script).exists():
                raise MissingItemError(f"script file {script} not found")

        def run(self, args, cwd=None, load_data=True):
            cmd = [str(a) for a in [self.exe, self.script, *args]]

            with contextlib.ExitStack() as stack:
                fpout = stack.enter_context((self.workdir / "stdout.txt").open("w"))
                fperr = stack.enter_context((self.workdir / "stderr.txt").open("w"))
                self.p = subprocess.Popen(
                    cmd,
                    cwd=self.workdir if cwd is True else cwd,
                    stdout=fpout,
                    stderr=fperr,
                )
                self.p.communicate()
            out = (self.workdir / "stdout.txt").read_text()
            err = (self.workdir / "stderr.txt").read_text()
            return Result(
                out.replace("\r\n", "\n"), err.replace("\r\n", "\n"), self.p.returncode
            )

        def compare(self, refdir, populate=False):
            src = self.datadir / refdir
            if not src.exists():
                raise MissingItemError(f"reference dir {src} not found")

            for name in ["stdout.txt", "stderr.txt"]:
                left = src / name
                right = self.workdir / name
                if populate:
                    if left.exists():
                        raise ScripterError(f"cannot overwrite {left} with {right}")
                    shutil.copyfile(right, left)
                assert left.read_text() == right.read_text()

    class Scripter:
        def __init__(self, srcdir, datadir, exe=sys.executable):
            self.srcdir = srcdir
            self.datadir = datadir
            self.exe = exe

        def __truediv__(self, path):
            tmpdir = tmp_path_factory.mktemp(pathlib.Path(path).with_suffix("").name)
            return Exe(self.srcdir / path, tmpdir, self.datadir, self.exe)

    return Scripter(pathlib.Path(request.module.__file__).parent, datadir)


@pytest.fixture(scope="function")
def git_project_factory(request, tmp_path):
    from pathlib import Path

    def indent(txt, pre=" " * 2):
        "simple text indentation"

        from textwrap import dedent

        txt = dedent(txt)
        if txt.endswith("\n"):
            last_eol = "\n"
            txt = txt[:-1]
        else:
            last_eol = ""

        return pre + txt.replace("\n", "\n" + pre) + last_eol

    class GitWrapper:
        EXE: str = "git"
        KEEPFILE = ".keep"

        def __init__(
            self,
            workdir: Union[Path, str],
            exe: Optional[str] = None,
            identity: Tuple[str, str] = ("First Last", "user@email"),
        ):
            self.workdir = Path(workdir)
            self.exe = exe or self.EXE
            self.identity = identity

        def init(
            self,
            force: bool = False,
            keepfile: Optional[Union[bool, Path]] = True,
            identity: Optional[Tuple[str, str]] = None,
        ) -> GitWrapper:
            from shutil import rmtree

            identity = self.identity if identity is None else identity
            keepfile = self.workdir / self.KEEPFILE if keepfile is True else keepfile

            if force:
                rmtree(self.workdir, ignore_errors=True)
            self.workdir.mkdir(parents=True, exist_ok=True if force else False)
            self("init")

            if identity:
                self(["config", "user.name", identity[0]])
                self(["config", "user.email", identity[1]])

            if keepfile:
                keepfile.write_text("# dummy file to create the master branch")
                self(["add", keepfile])
                self(["commit", "-m", "initial", keepfile])
            return self

        def clone(
            self,
            dest: Union[str, Path],
            force=False,
            branch: Optional[str] = None,
            exe: Optional[str] = None,
            identity: Optional[Tuple[str, str]] = None,
        ) -> GitWrapper:
            from shutil import rmtree

            dest = Path(dest)
            identity = self.identity if identity is None else identity
            exe = self.exe if exe is None else exe

            if force:
                rmtree(dest, ignore_errors=True)
            else:
                if dest.exists():
                    raise ValueError(f"target directory present {dest}")

            self(
                [
                    "clone",
                    *(["--branch", branch] if branch else []),
                    self.workdir.absolute(),
                    dest.absolute(),
                ],
            )

            result = self.__class__(dest, exe=exe, identity=identity)
            if identity:
                result(["config", "user.name", identity[0]])
                result(["config", "user.email", identity[1]])
            return result

        def __call__(self, cmd: Union[List[Any], Any], *args) -> str:
            arguments = []
            if isinstance(cmd, str):
                arguments.append(cmd)
            else:
                arguments.extend(cmd[:])

            if str(arguments[0]) != "clone":
                arguments = [
                    self.exe,
                    "--git-dir",
                    str(self.workdir.absolute() / ".git"),
                    "--work-tree",
                    str(self.workdir.absolute()),
                    *[str(a) for a in arguments],
                ]
            else:
                arguments = [self.exe, *[str(a) for a in arguments]]
            return subprocess.check_output(arguments, encoding="utf-8")

        def __truediv__(self, other):
            return self.workdir.absolute() / other

        def dump(self, fp: TextIO = sys.stdout) -> Optional[str]:
            lines = f"REPO: {self.workdir}"
            lines += "\n [status]\n" + indent(self(["status"]))
            lines += "\n [branch]\n" + indent(self(["branch", "-avv"]))
            lines += "\n [tags]\n" + indent(self(["tag", "-l"]))
            lines += "\n [remote]\n" + indent(self(["remote", "-v"]))
            if fp == io.StringIO:
                buf = io.StringIO()
                print(lines, file=buf)
                return buf.getvalue()
            else:
                print(lines, file=fp)
                return None

    class GitCli(GitWrapper):
        BETA_BRANCHES = re.compile(r"/beta/(?P<ver>\d+([.]\d+)*)")

        def commit(
            self, paths: Union[str, Path, List[Union[str, Path]]], message: str
        ) -> None:
            paths = [paths] if isinstance(paths, (Path, str)) else paths
            self(["add", *paths])
            self(["commit", "-m", message, *paths])

        def branch(self, name: Optional[str] = None, origin: str = "master") -> str:
            if not name:
                return self(["rev-parse", "--abbrev-ref", "HEAD"]).strip()
            assert origin or origin is None
            old = self.branch()
            self(["checkout", "-b", name, "--track", origin])
            return old

        def branches(
            self, expr: Optional[Union[str, bool, re.Pattern]] = None
        ) -> Tuple[List[str], Dict[str, List[str]]]:
            branches = self(["branch", "-a", "--format", "%(refname)"]).split()
            matchobj = re.compile(expr) if isinstance(expr, str) else expr

            if matchobj:
                branches = [name for name in branches if matchobj.search(name)]

            n = len("refs/heads/")
            local_branches = [
                name[n:] for name in branches if name.startswith("refs/heads/")
            ]
            remote_branches: Dict[str, List[str]] = {}
            n = len("refs/remotes/")
            for name in branches:
                if not name.startswith("refs/remotes/"):
                    continue
                origin, _, name = name[n:].partition("/")
                if origin not in remote_branches:
                    remote_branches[origin] = []
                remote_branches[origin].append(name)
            return local_branches, remote_branches

    class Project(GitCli):
        @property
        def initfile(self):
            return self.workdir / "src" / "__init__.py"

        @property
        def version(self):
            return (
                self.initfile.read_text()
                .partition("=")[2]
                .strip()
                .replace("'", '"')
                .strip('"')
            )

        def create(self, version=None, clone=None, force=False):
            # repo = git_project_factory().create("0.0.0")
            # repo1 = git_project_factory().create(clone=repo)
            if clone:
                clone.clone(self.workdir, force=force)
            else:
                self.init(force=force)

            if version is not None:
                self.initfile.parent.mkdir(parents=True, exist_ok=True)
                self.initfile.write_text(
                    f"""
    __version__ = "{version}"
""".lstrip()
                )
                self.commit([self.initfile], "initial commit")
            return self

    return lambda subdir="": Project(tmp_path / (subdir or request.node.name))


def pytest_configure(config):
    config.addinivalue_line("markers", "manual: test intented to run manually")


def pytest_collection_modifyitems(config, items):
    if config.option.keyword or config.option.markexpr:
        return  # let pytest handle this

    for item in items:
        if "manual" not in item.keywords:
            continue
        item.add_marker(pytest.mark.skip(reason="manual not selected"))



import re
from pathlib import Path
from typing import List, Callable, Dict, TypeVar

from mypy_extensions import Arg, DefaultArg

from . import tools

BETAEXPR = re.compile(r"beta/(?P<ver>\d+([.]\d+)*)")

ErrorFunctionType = TypeVar(
    "ErrorFunctionType",
    bound=Callable[
        [
            Arg(str, "message"),  # noqa: undefined-name
            DefaultArg(str, "explain"),  # noqa: undefined-name
            DefaultArg(str, "hint"),  # noqa: undefined-name
        ],
        None,
    ],
)


def check_repo_mods(error: ErrorFunctionType, workdir: Path, initfile: Path):
    from pygit2 import Repository, GIT_STATUS_WT_NEW, GIT_STATUS_IGNORED  # type: ignore

    repo = Repository(workdir)

    rel_initfile = str(initfile.absolute().relative_to(workdir.absolute())).replace(
        "\\", "/"
    )
    unstracked = {p for p, f in repo.status().items() if f & GIT_STATUS_WT_NEW}
    if rel_initfile in unstracked:
        error(
            "init file is not tracked",
            explain="""
                  An init file (eg. __init__.py) should be defined containing
                  a __version__ = "<major>.<minor>.<micro>" version
                  """,
            hint=f"create and git add an init file in '{initfile}'",
        )

    def ignore(f):
        return (f & GIT_STATUS_WT_NEW) or (f & GIT_STATUS_IGNORED)

    modified = {p for p, f in repo.status().items() if not ignore(f)}
    if rel_initfile in modified:
        error(
            "init file has local modifications",
            explain="""
                  An init file (eg. __init__.py) should be git tracked and not modified
                  """,
            hint=f"git revert or commit init file changes in '{initfile}'",
        )


def check_initfile(error: ErrorFunctionType, initfile: Path) -> None:
    if not initfile.exists():
        error(
            "no init file found",
            explain="""
                  An init file (eg. __init__.py) should be defined containing
                  a __version__ = "<major>.<minor>.<micro>" version
                  """,
            hint=f"add an init file in '{initfile}'",
        )
    curver = tools.get_module_var(initfile, "__version__", abort=False)
    if not curver:
        error(
            "init file has an invalid __version__ module variable",
            explain="""
          An init file (eg. __init__.py) should be defined containing
          a __version__ = "<major>.<minor>.<micro>" version
          """,
            hint=f"add a __version__ module variable in '{initfile}'",
        )


def check_branch(
    error: ErrorFunctionType,
    mode: str,
    curbranch: str,
    master: str = "master",
):
    # curbranch == repo.head.shorthand
    if mode in {"release"}:
        match = BETAEXPR.search(curbranch)
        if not match:
            error(
                f"{mode} starts from a beta/N.M.O branch",
                f"""
                A {mode} starts from a beta/N.M.O branch, not from '{curbranch}'
                """,
                hint="switch to a beta/N.M.O branch",
            )
    elif mode in {"major", "minor", "micro"}:
        # betas start from the 'master' branch
        if curbranch != master:
            error(
                f"'{mode}' starts from '{master}' branch",
                f"""
                While generating a branch for '{mode}' we assume as starting
                branch to be '{master}' but we are in '{curbranch}'.
                """,
                hint=f"""
                Switch to the '{master}' branch or pass the --master flag
                """,
            )
    else:
        raise RuntimeError(f"invalid {mode}")


def check_version(
    error: ErrorFunctionType,
    mode: str,
    initfile: Path,
    branch: str,
    local_branches: List[str],
    remote_branches: Dict[str, List[str]],
    tags: List[str],
    master: str,
):
    curver = tools.get_module_var(initfile, "__version__", abort=False)
    nextver = tools.bump_version(curver or "", mode)

    if mode in {"release"}:
        if branch != f"beta/{curver}":
            error(
                f"wrong version '{curver}' from initfile",
                f"""
                The current branch '{branch}' has a version '{curver}' in
                the initfile '{initfile}'.
                """,
                hint="""
                fix the __version__ variable in the initfile
                """,
            )

        if f"release/{curver}" in tags:
            error(
                "release already prsent",
                f"""
                A release 'release/{curver}' tag is present for the current branch
                """,
                hint="""
                check the __version__ is correct
                """,
            )
    else:
        if f"beta/{nextver}" in local_branches:
            error(
                f"next version branch 'beta/{nextver}' already present"
                " in local branches",
                f"""
                when creating a new branch 'beta/{nextver}' a local branch
                with that name has been found already
                """,
                hint=f"""
                change the version from '{curver}' in the '{master}' branch initfile
                """,
            )
        for origin, branches in remote_branches.items():
            if f"beta/{nextver}" in branches:
                error(
                    f"next version branch 'beta/{nextver}' already present in"
                    " remote branches",
                    f"""
                when creating a new branch 'beta/{nextver}' a remote branch with
                that name has been found already in '{origin}'
                """,
                    hint=f"""
                make sure the '{curver}' in the initfile in '{master}' branch is correct
                """,
                )



import pytest

from setuptools_github import checks, tools


def error(message: str, explain: str = "", hint: str = ""):
    raise tools.AbortExecution(message, explain, hint)


def test_check_initfile(git_project_factory):
    project = git_project_factory().create()

    exc = pytest.raises(
        tools.AbortExecution, checks.check_initfile, error, project.initfile
    ).value
    assert """
no init file found
  An init file (eg. __init__.py) should be defined containing
  a __version__ = "<major>.<minor>.<micro>" version
hint:
  add an init file in""".strip() in str(
        exc
    )

    # add an empty init file
    project.initfile.parent.mkdir(parents=True)
    project.initfile.write_text("hello =  1\n")

    exc = pytest.raises(
        tools.AbortExecution, checks.check_initfile, error, project.initfile
    ).value
    assert """
init file has an invalid __version__ module variable
  An init file (eg. __init__.py) should be defined containing
  a __version__ = "<major>.<minor>.<micro>" version
hint:
  add a __version__ module variable in""".strip() in str(
        exc
    )


def test_check_branch(git_project_factory):
    project = git_project_factory().create("0.0.3")
    assert project.branch() == "master"

    pytest.raises(
        RuntimeError, checks.check_branch, error, "no-value", project.branch()
    )

    # switch branch
    old = project.branch("abc")
    assert (old, project.branch()) == ("master", "abc")

    # check we cannot start from that branch
    exc = pytest.raises(
        tools.AbortExecution, checks.check_branch, error, "minor", project.branch()
    ).value
    assert """
'minor' starts from 'master' branch
  While generating a branch for 'minor' we assume as starting
  branch to be 'master' but we are in 'abc'.
hint:
  Switch to the 'master' branch or pass the --master flag
""".strip() in str(
        exc
    )

    # we force the master to be 'abc'
    assert not checks.check_branch(error, "minor", project.branch(), master="abc")

    # starting a release branch
    exc = pytest.raises(
        tools.AbortExecution,
        checks.check_branch,
        error,
        "release",
        project.branch(),
        master="abc",
    ).value
    assert """
release starts from a beta/N.M.O branch
  A release starts from a beta/N.M.O branch, not from 'abc'
hint:
  switch to a beta/N.M.O branch
""".strip() in str(
        exc
    )

    project.branch("beta/1.2.3")
    assert not checks.check_branch(error, "release", project.branch(), master="abc")


def test_check_version(git_project_factory):
    from setuptools_github.start_release import extract_beta_branches

    repo = git_project_factory("test_check_version-repo").create("0.0.0")
    repo1 = git_project_factory("test_check_version-repo1").create(clone=repo)

    repo.branch("beta/0.0.3")
    repo(["tag", "-m", "release", "release/0.0.3"])
    repo.branch("beta/0.0.4")
    repo(["tag", "-m", "release", "release/0.0.4"])
    repo1.branch("beta/0.0.2")

    project = git_project_factory().create(clone=repo)
    project.branch("beta/0.0.1", "origin/master")
    project.branch("master", "origin/master")

    project(["remote", "add", "repo1", repo1.workdir])
    project(["fetch", "--all"])

    local_branches, remote_branches, tags = [
        *project.branches(project.BETA_BRANCHES),
        project(["tag", "-l"]).split(),
    ]
    from pygit2 import Repository

    repo = Repository(project.workdir)
    assert (local_branches, remote_branches, tags) == extract_beta_branches(repo)

    assert project.branch() == "master"
    assert project.version == "0.0.0"

    exc = pytest.raises(
        tools.AbortExecution,
        checks.check_version,
        error,
        "micro",
        project.initfile,
        project.branch(),
        local_branches,
        remote_branches,
        tags,
        "master",
    ).value
    assert """
next version branch 'beta/0.0.1' already present in local branches
  when creating a new branch 'beta/0.0.1' a local branch
  with that name has been found already
hint:
  change the version from '0.0.0' in the 'master' branch initfile
""".strip() in str(
        exc
    )

    project.initfile.write_text("__version__ = '0.0.1'")
    project.commit(project.initfile, "update")
    exc = pytest.raises(
        tools.AbortExecution,
        checks.check_version,
        error,
        "micro",
        project.initfile,
        project.branch(),
        local_branches,
        remote_branches,
        tags,
        "master",
    ).value
    assert """
next version branch 'beta/0.0.2' already present in remote branches
  when creating a new branch 'beta/0.0.2' a remote branch with
  that name has been found already in 'repo1'
hint:
  make sure the '0.0.1' in the initfile in 'master' branch is correct
""".strip() in str(
        exc
    )

    project.initfile.write_text("__version__ = '0.0.4'")
    project.commit(project.initfile, "update")
    assert not checks.check_version(
        error,
        "micro",
        project.initfile,
        project.branch(),
        local_branches,
        remote_branches,
        tags,
        "master",
    )

    # release checks
    project(["checkout", "beta/0.0.4"])
    project(["merge", "master"])
    assert (project.branch(), project.version) == ("beta/0.0.4", "0.0.4")

    exc = pytest.raises(
        tools.AbortExecution,
        checks.check_version,
        error,
        "release",
        project.initfile,
        project.branch(),
        local_branches,
        remote_branches,
        tags,
        "master",
    ).value
    assert """
release already prsent
  A release 'release/0.0.4' tag is present for the current branch
hint:
  check the __version__ is correct
""".strip() in str(
        exc
    )



from __future__ import annotations

import io
import re
import ast
import sys
import json
import subprocess

from pathlib import Path
from typing import Any, Union, Tuple, Optional, List, Dict, TextIO


class GithubError(Exception):
    pass


class MissingVariable(GithubError):
    pass


class InvalidGithubReference(GithubError):
    pass


class AbortExecution(Exception):
    @staticmethod
    def _strip(txt):
        txt = txt or ""
        txt = txt[1:] if txt.startswith("\n") else txt
        txt = indent(txt, pre="")
        return txt[:-1] if txt.endswith("\n") else txt

    def __init__(
        self, message: str, explain: Optional[str] = None, hint: Optional[str] = None
    ):
        self.message = message.strip()
        self._explain = explain
        self._hint = hint

    @property
    def explain(self):
        return self._strip(self._explain)

    @property
    def hint(self):
        return self._strip(self._hint)

    def __str__(self):
        result = [self.message]
        if self.explain:
            result.append(indent("\n" + self.explain, pre=" " * 2)[2:])
        if self.hint:
            result.extend(["\nhint:", indent("\n" + self.hint, pre=" " * 2)[2:]])
        return "".join(result)


def urmtree(path: Path):
    from os import name
    from shutil import rmtree
    from stat import S_IWUSR
    if name == "nt":
        for p in path.rglob("*"):
            p.chmod(S_IWUSR)
    rmtree(path, ignore_errors=True)
    assert not path.exists()


def indent(txt: str, pre: str = " " * 2) -> str:
    "simple text indentation"

    from textwrap import dedent

    txt = dedent(txt)
    if txt.endswith("\n"):
        last_eol = "\n"
        txt = txt[:-1]
    else:
        last_eol = ""

    return pre + txt.replace("\n", "\n" + pre) + last_eol


def hubversion(gdata: Any, fallback: Optional[str]) -> Tuple[Optional[str], str]:
    """gets a (version, sha) tuple from gdata.

    GITHUB_DUMP is a json dump of the environment during an action run and
    we pull the ref, run_number and sha keys.

    ref can be something like refs/heads/master for the master branch,
    refs/heads/beta/0.0.4 for a beta branch or refs/tags/release/0.0.3 for
    a release tag.

    the version returned is:
        ("", "<sha-value>") for the master branch
        ("0.0.4b8", "<sha-value>") for a beta branch (<version>b<build-number>)
        ("0.0.3", "<sha-value>") for the release

    Args:
        gdata: json dictionary from $GITHUB_DUMP
        fallback: returns this if a version is not defined in gdata

    Returns:
        (str, str): <update-version>, <shasum>
    """

    def validate(txt):
        return ".".join(str(int(v)) for v in txt.split("."))

    ref = gdata["ref"]  # eg. "refs/tags/release/0.0.3"
    number = gdata["run_number"]  # eg. 3
    shasum = gdata["sha"]  # eg. "2169f90c"

    # the logic for the returned version:

    # 1. if we are on master we use the version from the __init__.py module
    if ref == "refs/heads/master":
        return (fallback, shasum)

    # 2. on a beta branch we add a "b<build-number>" string to the __init__.py version
    #    the bersion is taken from the refs/heads/beta/<version>
    if ref.startswith("refs/heads/beta/"):
        version = validate(ref.rpartition("/")[2])
        return (f"{version}b{number}", shasum)

    # 3. on a release we use the version from the refs/tags/release/<version>
    if ref.startswith("refs/tags/release/"):
        version = validate(ref.rpartition("/")[2])
        return (f"{version}", shasum)

    raise InvalidGithubReference("unhandled github ref", gdata)


def get_module_var(
    path: Union[Path, str], var: str = "__version__", abort=True
) -> Optional[str]:
    """extract from a python module in path the module level <var> variable

    Args:
        path (str,Path): python module file to parse using ast (no code-execution)
        var (str): module level variable name to extract
        abort (bool): raise MissingVariable if var is not present

    Returns:
        None or str: the variable value if found or None

    Raises:
        MissingVariable: if the var is not found and abort is True

    Notes:
        this uses ast to parse path, so it doesn't load the module
    """

    class V(ast.NodeVisitor):
        def __init__(self, keys):
            self.keys = keys
            self.result = {}

        def visit_Module(self, node):
            # we extract the module level variables
            for subnode in ast.iter_child_nodes(node):
                if not isinstance(subnode, ast.Assign):
                    continue
                for target in subnode.targets:
                    if target.id not in self.keys:
                        continue
                    assert isinstance(
                        subnode.value, (ast.Num, ast.Str, ast.Constant)
                    ), (
                        f"cannot extract non Constant variable "
                        f"{target.id} ({type(subnode.value)})"
                    )
                    if isinstance(subnode.value, ast.Str):
                        value = subnode.value.s
                    elif isinstance(subnode.value, ast.Num):
                        value = subnode.value.n
                    else:
                        value = subnode.value.value
                    self.result[target.id] = value
            return self.generic_visit(node)

    v = V({var})
    path = Path(path)
    if path.exists():
        tree = ast.parse(Path(path).read_text())
        v.visit(tree)
    if var not in v.result and abort:
        raise MissingVariable(f"cannot find {var} in {path}", path, var)
    return v.result.get(var, None)


def set_module_var(
    path: Union[str, Path], var: str, value: Any, create: bool = True
) -> Tuple[Any, str]:
    """replace var in path with value

    Args:
        path (str,Path): python module file to parse
        var (str): module level variable name to extract
        value (None or Any): if not None replace var in initfile
        create (bool): create path if not present

    Returns:
        (str, str) the (<previous-var-value|None>, <the new text>)
    """
    # module level var
    expr = re.compile(f"^{var}\\s*=\\s*['\\\"](?P<value>[^\\\"']*)['\\\"]")
    fixed = None
    lines = []

    src = Path(path)
    if not src.exists() and create:
        src.parent.mkdir(parents=True, exist_ok=True)
        src.touch()

    input_lines = src.read_text().split("\n")
    for line in reversed(input_lines):
        if fixed:
            lines.append(line)
            continue
        match = expr.search(line)
        if match:
            fixed = match.group("value")
            if value is not None:
                x, y = match.span(1)
                line = line[:x] + value + line[y:]
        lines.append(line)
    txt = "\n".join(reversed(lines))
    if not fixed and create:
        if txt and txt[-1] != "\n":
            txt += "\n"
        txt += f'{var} = "{value}"'

    with Path(path).open("w") as fp:
        fp.write(txt)
    return fixed, txt


def update_version(
    initfile: Union[str, Path], github_dump: Optional[str] = None
) -> Optional[str]:
    """extracts version information from github_dump and updates initfile in-place

    Args:
        initfile (str, Path): path to the __init__.py file with a __version__ variable
        github_dump (str): the os.getenv("GITHUB_DUMP") value

    Returns:
        str: the new version for the package
    """

    path = Path(initfile)

    if not github_dump:
        return get_module_var(path, "__version__")
    gdata = json.loads(github_dump) if isinstance(github_dump, str) else github_dump

    version, thehash = hubversion(gdata, get_module_var(path, "__version__"))
    set_module_var(path, "__version__", version)
    set_module_var(path, "__hash__", thehash)
    return version


def bump_version(version: str, mode: str) -> str:
    """given a version str will bump it according to mode

    Arguments:
        version: text in the N.M.O form
        mode: major, minor or micro

    Returns:
        increased text

    >>> bump_version("1.0.3", "micro")
    "1.0.4"
    >>> bump_version("1.0.3", "minor")
    "1.1.0"
    """
    newver = [int(n) for n in version.split(".")]
    if mode == "major":
        newver[-3] += 1
        newver[-2] = 0
        newver[-1] = 0
    elif mode == "minor":
        newver[-2] += 1
        newver[-1] = 0
    elif mode == "micro":
        newver[-1] += 1
    return ".".join(str(v) for v in newver)


class GitWrapper:
    EXE: str = "git"
    KEEPFILE = ".keep"

    def __init__(
        self,
        workdir: Union[Path, str],
        exe: Optional[str] = None,
        identity: Tuple[str, str] = ("First Last", "user@email"),
    ):
        self.workdir = Path(workdir)
        self.exe = exe or self.EXE
        self.identity = identity

    def init(
        self,
        force: bool = False,
        keepfile: Optional[Union[bool, Path]] = True,
        identity: Optional[Tuple[str, str]] = None,
    ) -> GitWrapper:

        identity = self.identity if identity is None else identity
        keepfile = self.workdir / self.KEEPFILE if keepfile is True else keepfile

        if force:
            urmtree(self.workdir)
        self.workdir.mkdir(parents=True, exist_ok=True if force else False)
        self("init")

        if identity:
            self(["config", "user.name", identity[0]])
            self(["config", "user.email", identity[1]])

        if keepfile:
            keepfile.write_text("# dummy file to create the master branch")
            self(["add", keepfile])
            self(["commit", "-m", "initial", keepfile])
        return self

    def clone(
        self,
        dest: Union[str, Path],
        force=False,
        branch: Optional[str] = None,
        exe: Optional[str] = None,
        identity: Optional[Tuple[str, str]] = None,
    ) -> GitWrapper:
        dest = Path(dest)
        identity = self.identity if identity is None else identity
        exe = self.exe if exe is None else exe

        if force:
            urmtree(dest)
        else:
            if dest.exists():
                raise ValueError(f"target directory present {dest}")

        self(
            [
                "clone",
                *(["--branch", branch] if branch else []),
                self.workdir.absolute(),
                dest.absolute(),
            ],
        )

        result = self.__class__(dest, exe=exe, identity=identity)
        if identity:
            result(["config", "user.name", identity[0]])
            result(["config", "user.email", identity[1]])
        return result

    def __call__(self, cmd: Union[List[Any], Any], *args) -> str:
        arguments = []
        if isinstance(cmd, str):
            arguments.append(cmd)
        else:
            arguments.extend(cmd[:])

        if str(arguments[0]) != "clone":
            arguments = [
                self.exe,
                "--git-dir",
                str(self.workdir.absolute() / ".git"),
                "--work-tree",
                str(self.workdir.absolute()),
                *[str(a) for a in arguments],
            ]
        else:
            arguments = [self.exe, *[str(a) for a in arguments]]
        return subprocess.check_output(arguments, encoding="utf-8")

    def __truediv__(self, other):
        return self.workdir.absolute() / other

    def dump(self, fp: TextIO = sys.stdout) -> Optional[str]:
        lines = f"REPO: {self.workdir}"
        lines += "\n [status]\n" + indent(self(["status"]))
        lines += "\n [branch]\n" + indent(self(["branch", "-avv"]))
        lines += "\n [tags]\n" + indent(self(["tag", "-l"]))
        lines += "\n [remote]\n" + indent(self(["remote", "-v"]))
        if fp == io.StringIO:
            buf = io.StringIO()
            print(lines, file=buf)
            return buf.getvalue()
        else:
            print(lines, file=fp)
            return None


class GitCli(GitWrapper):
    BETA_BRANCHES = re.compile(r"/beta/(?P<ver>\d+([.]\d+)*)")

    def commit(
        self, paths: Union[str, Path, List[Union[str, Path]]], message: str
    ) -> None:
        paths = [paths] if isinstance(paths, (Path, str)) else paths
        self(["add", *paths])
        self(["commit", "-m", message, *paths])

    def branch(self, name: Optional[str] = None, origin: str = "master") -> str:
        if not name:
            return self(["rev-parse", "--abbrev-ref", "HEAD"]).strip()
        assert origin or origin is None
        old = self.branch()
        self(["checkout", "-b", name, "--track", origin])
        return old

    def branches(
        self, expr: Optional[Union[str, re.Pattern]] = None
    ) -> Tuple[List[str], Dict[str, List[str]]]:
        branches = self(["branch", "-a", "--format", "%(refname)"]).split()
        matchobj = re.compile(expr) if isinstance(expr, str) else expr

        if matchobj:
            branches = [name for name in branches if matchobj.search(name)]

        n = len("refs/heads/")
        local_branches = [
            name[n:] for name in branches if name.startswith("refs/heads/")
        ]
        remote_branches: Dict[str, List[str]] = {}
        n = len("refs/remotes/")
        for name in branches:
            if not name.startswith("refs/remotes/"):
                continue
            origin, _, name = name[n:].partition("/")
            if origin not in remote_branches:
                remote_branches[origin] = []
            remote_branches[origin].append(name)
        return local_branches, remote_branches



import argparse
import sys
import logging
from pathlib import Path
import functools
import re
from typing import List, Optional, Tuple, Dict, Any

import pygit2  # type: ignore

from setuptools_github import tools
from setuptools_github import checks


log = logging.getLogger(__name__)


def parse_args(args: Optional[str] = None, testmode: bool = False) -> Dict[str, Any]:
    """parses args from the command line

    Args:
        args: command line arguments or None to pull from sys.argv
        testmode: internal flag, if set will not SystemExit but will
                  raises tools.AbortExecution
    """
    from pathlib import Path

    class F(
        argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter
    ):
        pass

    parser = argparse.ArgumentParser(
        formatter_class=F, description=__doc__, prog="AAAA"
    )

    parser.add_argument("-n", "--dry-run", dest="dryrun", action="store_true")
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("--master", default="master", help="the 'master' branch")

    # parser.add_argument("-f", "--force", action="store_true")
    # parser.add_argument("--no-checks", action="store_true")

    parser.add_argument(
        "-w",
        "--workdir",
        help="git working dir",
        default=Path("."),
        type=Path,
    )
    parser.add_argument("mode", choices=["micro", "minor", "major", "release"])
    parser.add_argument("initfile", metavar="__init__.py", type=Path)

    options = parser.parse_args(args)

    def error(message, explain="", hint="", parser=None, testmode=False):
        out = []
        if parser:
            out.extend(tools.indent(parser.format_usage()).split("\n"))
        if message:
            out.extend(tools.indent(message).split("\n"))
        if explain:
            out.append("reason:")
            out.extend(tools.indent(explain).split("\n"))
        if hint:
            out.append("hint:")
            out.extend(tools.indent(hint).split("\n"))

        if testmode:
            raise tools.AbortExecution(message, explain, hint)
        else:
            print()
            print("\n".join(out), file=sys.stderr)
            raise SystemExit(2)

    options.error = functools.partial(error, parser=parser, testmode=testmode)

    logging.basicConfig(
        format="%(levelname)s:%(name)s:(dry-run) %(message)s"
        if options.dryrun
        else "%(levelname)s:%(name)s:%(message)s",
        level=logging.DEBUG if options.verbose else logging.INFO,
    )

    for d in [
        "verbose",
    ]:
        delattr(options, d)
    return options.__dict__


def extract_beta_branches(
    repo: pygit2.Repository,
) -> Tuple[List[str], Dict[str, List[str]], List[str]]:
    """given a pygit2 Repository object extracts local and remote beta branches

     This function will extract all the 'beta' branches (eg. with the beta/N(.N)* form)
     and the release tags (eg. with release/N(.N)* form) from it.

    Examples:
         >>> extract_beta_branches(pygit2.Repository(.. some path))
         (
             ['beta/0.0.1', 'beta/0.0.4'],
             {'origin': ['beta/0.0.3', 'beta/0.0.4'], 'repo1': ['beta/0.0.2']},
             ['release/0.0.3', 'release/0.0.4']
         )
    """
    tagre = re.compile(r"^refs/tags/release/")
    local_branches = []
    remote_branches: Dict[str, List[str]] = {}
    for name in repo.branches.local:
        if checks.BETAEXPR.search(name):
            local_branches.append(name)

    for name in repo.branches.remote:
        if checks.BETAEXPR.search(name):
            origin, _, name = name.partition("/")
            if origin not in remote_branches:
                remote_branches[origin] = []
            remote_branches[origin].append(name)

    pre = len("refs/tags/")
    tags = [name[pre:] for name in repo.references if tagre.search(name)]
    return local_branches, remote_branches, tags


def beta(
    repo: pygit2.Repository,
    curver: str,
    mode: str,
    initfile: Path,
    workdir: Path,
    dryrun: bool,
    error: checks.ErrorFunctionType,
) -> None:
    newver = tools.bump_version(curver or "", mode)
    log.info("beta release %s -> %s", curver, newver)

    local_branches, remote_branches, tags = extract_beta_branches(repo)

    # it's the first branch, we won't bump the version
    if not (local_branches or remote_branches):
        newver = curver
    newbranch = f"beta/{newver}"

    if newver == curver:
        log.info(
            "creating first version branch '%s' (v. %s) from 'master'",
            newbranch,
            newver,
        )
    else:
        log.info(
            "creating new version branch '%s' (v. %s) from 'master' (%s)",
            newbranch,
            newver,
            curver,
        )

    # modify the __init__
    log.info("updating init file %s (%s -> %s)", initfile, curver, newver)
    if not dryrun:
        tools.set_module_var(initfile, "__version__", newver)

    # commit the updated __init__.py in the master branch
    msg = f"beta release {newver}"
    log.info("committing '%s'%s", msg, " (skip)" if dryrun else "")
    if not dryrun:
        refname = repo.head.name
        author = repo.default_signature
        commiter = repo.default_signature
        parent = repo.revparse_single(repo.head.shorthand).hex
        relpath = initfile.absolute().relative_to(workdir.absolute())
        repo.index.add(str(relpath).replace("\\", "/"))
        repo.index.write()
        tree = repo.index.write_tree()
        oid = repo.create_commit(refname, author, commiter, msg, tree, [parent])
        log.info("created oid %s", oid)

    # log.info("switching to new branch '%s'%s", newbranch, " (skip)" if dryrun else "")
    if not dryrun:
        commit = repo.revparse_single(repo.head.shorthand)
        repo.branches.local.create(newbranch, commit)
        ref = repo.lookup_reference(repo.lookup_branch(newbranch).name)
        repo.checkout(ref)


def release(
    repo: pygit2.Repository,
    curver: str,
    mode: str,
    initfile: Path,
    workdir: Path,
    dryrun: bool,
    error: checks.ErrorFunctionType,
) -> None:
    log.info("releasing %s", curver)

    local_branches, remote_branches, tags = extract_beta_branches(repo)

    ref = None
    if f"beta/{curver}" in local_branches:
        ref = repo.lookup_reference(f"refs/heads/beta/{curver}")
    else:
        # TODO check local and remote branches aren't out of sync
        for origin, branches in remote_branches.items():
            if f"beta/{curver}" not in branches:
                continue

            ref = repo.lookup_reference(f"refs/remotes/{origin}/beta/{curver}")
            break

    if not ref:
        error(
            "cannot find a suitable branch/ref",
            """cannot find a reference for beta/{curver}
              """,
        )
    elif not dryrun:
        repo.checkout(ref)
        repo.references.create(f"refs/tags/release/{curver}", ref.target)


def run(mode, initfile, workdir, dryrun, error, master):
    workdir = workdir.resolve()
    log.debug("using working dir %s", workdir)

    # get the Repository instance on workdir
    repo = pygit2.Repository(workdir)
    try:
        branch = repo.head.shorthand
    except pygit2.GitError:
        error(
            "invalid git repository",
            """
              It looks the repository doesn't have any branch,
              you should:
                git checkout --orphan <branch-name>
              """,
            hint="create a git branch",
        )

    # get all the beta branches
    # check the current version has a beta/<curver> branch

    # TODO check for local modifications
    checks.check_repo_mods(error, workdir, initfile)
    checks.check_initfile(error, initfile)
    checks.check_branch(error, mode, branch, master)

    # TODO check for branch matching init file
    local_branches, remote_branches, tags = extract_beta_branches(repo)
    checks.check_version(
        error, mode, initfile, branch, local_branches, remote_branches, tags, branch
    )

    curver = tools.get_module_var(initfile, "__version__", abort=False)
    log.info("current version [%s]", curver)
    return (release if mode == "release" else beta)(
        repo, curver, mode, initfile, workdir, dryrun, error
    )

    return

    # get the current version from initfile
    curver = tools.get_module_var(initfile, "__version__", abort=False)
    if not curver:
        error(
            f"cannot find __version__ in {initfile}",
            explain="""
        The initfile should contain the __version__ module level variable;
        it should be a text string in the MAJOR.MINOR.MICRO form.
        """,
        )
    log.info("current version [%s]", curver)

    return (release if mode == "release" else beta)(
        repo, curver, mode, initfile, workdir, dryrun, error
    )


def main():
    run(**parse_args())


if __name__ == "__main__":
    main()
